#ifndef RUBY_H                                       /*-*-C++-*-vi:se ft=cpp:*/
#define RUBY_H 1
/**
 * @author     $Author$
 * @date       Sun 10 12:06:15 Jun JST 2007
 * @copyright  2007-2008 Yukihiro Matsumoto
 * @copyright  This  file  is   a  part  of  the   programming  language  Ruby.
 *             Permission  is hereby  granted,  to  either redistribute  and/or
 *             modify this file, provided that  the conditions mentioned in the
 *             file COPYING are met.  Consult the file for details.
 */
#define HAVE_RUBY_ATOMIC_H      1
#define HAVE_RUBY_DEBUG_H       1
#define HAVE_RUBY_DEFINES_H     1
#define HAVE_RUBY_ENCODING_H    1
#define HAVE_RUBY_FIBER_SCHEDULER_H 1
#define HAVE_RUBY_INTERN_H      1
#define HAVE_RUBY_IO_H          1
#define HAVE_RUBY_MEMORY_VIEW_H 1
#define HAVE_RUBY_MISSING_H     1
#define HAVE_RUBY_ONIGMO_H      1
#define HAVE_RUBY_ONIGURUMA_H   1
#define HAVE_RUBY_RACTOR_H      1
#define HAVE_RUBY_RANDOM_H      1
#define HAVE_RUBY_RE_H          1
#define HAVE_RUBY_REGEX_H       1
#define HAVE_RUBY_RUBY_H        1
#define HAVE_RUBY_ST_H          1
#define HAVE_RUBY_THREAD_H      1
#define HAVE_RUBY_THREAD_NATIVE_H 1
#define HAVE_RUBY_UTIL_H        1
#define HAVE_RUBY_VERSION_H     1
#define HAVE_RUBY_VM_H          1
#ifdef _WIN32
#define HAVE_RUBY_WIN32_H       1
#endif

#include "ruby/ruby.h"

#endif /* RUBY_H */
