#ifndef EXTCONF_H
#define EXTCONF_H
#define HAVE_BUILTIN___BUILTIN_CLZ 1
#define HAVE_BUILTIN___BUILTIN_CLZL 1
#define HAVE_BUILTIN___BUILTIN_CLZLL 1
#define HAVE_FLOAT_H 1
#define HAVE_MATH_H 1
#define HAVE_STDBOOL_H 1
#define HAVE_STDLIB_H 1
#define HAVE_X86INTRIN_H 1
#define HAVE_RUBY_ATOMIC_H 1
#define HAVE_RUBY_INTERNAL_HAS_BUILTIN_H 1
#define HAVE_RUBY_INTERNAL_STATIC_ASSERT_H 1
#define HAVE_RB_COMPLEX_REAL 1
#define HAVE_RB_COMPLEX_IMAG 1
#define HAVE_RB_OPTS_EXCEPTION_P 1
#define HAVE_RB_CATEGORY_WARN 1
#define HAVE_CONST_RB_WARN_CATEGORY_DEPRECATED 1
#endif
