#ifndef EXTCONF_H
#define EXTCONF_H
#define HAVE_FFI_H 1
#define RUBY_LIBFFI_MODVERSION 3005002
#define USE_FFI_CLOSURE_ALLOC 1
#define HAVE_FFI_PREP_CIF_VAR 1
#define HAVE_SYS_MMAN_H 1
#define HAVE_LINK_H 1
#define HAVE_DLFCN_H 1
#define HAVE_DLOPEN 1
#define HAVE_DLCLOSE 1
#define HAVE_DLSYM 1
#define HAVE_DLERROR 1
#define HAVE_DLINFO 1
#define HAVE_CONST_RTLD_DI_LINKMAP 1
#define SIGNEDNESS_OF_SIZE_T +1
#define HAVE_RB_STR_TO_INTERNED_STR 1
#endif
